import numpy as np
import pytest
import sasktran2 as sk


def test_dbm():
    """
    Tests the DBM optical property against a few hardcoded test results obtained from SASKTRAN legacy
    """
    dbm = sk.optical.O3DBM()

    config = sk.Config()

    altitude_grid = np.arange(0, 65001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    # Old SASKTRAN test results
    test_cases = []
    test_cases.append(
        {
            "temperature": 240,
            "values": np.array(
                [8.71858864e-20, 3.03768968e-21, 1.03810358e-22, 5.20901605e-21]
            ),
        }
    )
    test_cases.append(
        {
            "temperature": 228,
            "values": np.array(
                [8.47813970e-20, 2.78021858e-21, 7.71502986e-23, 5.22090221e-21]
            ),
        }
    )

    # Old SASKTRAN wavelengths are messed up
    wavel = sk.optical.air_wavelength_to_vacuum_wavelength(
        np.array([310, 330, 350, 600])
    )

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    for test_case in test_cases:
        atmosphere.temperature_k[:] = test_case["temperature"]

        result = dbm.atmosphere_quantities(atmo=atmosphere).extinction[10, :]

        np.testing.assert_allclose(result, test_case["values"] * 1e-4)


def test_o3_hitran():
    """
    Checks O3DBM against O3 HITRAN.  They have significant differences so we check that this does fail.
    """
    dbm = sk.optical.O3DBM()
    hitran = sk.optical.HITRANUV("o3")

    config = sk.Config()

    altitude_grid = np.arange(0, 65001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    wavel = np.arange(250, 350, 0.001)

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    xs_dbm = dbm.atmosphere_quantities(atmo=atmosphere).extinction[10, :]
    xs_hitran = hitran.atmosphere_quantities(atmo=atmosphere).extinction[10, :]

    with pytest.raises(AssertionError):
        np.testing.assert_allclose(xs_dbm, xs_hitran)



import sasktran2 as sk


def test_viewing_container():
    """
    Tests that we can create each different type of viewing ray and add them to the container without issue
    """
    viewing_geo = sk.ViewingGeometry()

    ray = sk.TangentAltitudeSolar(10000, 0, 600000, 0.6)

    viewing_geo.add_ray(ray)

    ray = sk.GroundViewingSolar(0.6, 0, -0.8, 200000)

    viewing_geo.add_ray(ray)

    assert len(viewing_geo.observer_rays) == 2



import numpy as np
import sasktran2 as sk
from scipy.special import eval_legendre


def test_wigner_construction():
    _ = sk.util.WignerD(0, 0)


def test_wigner_vectorization():
    wd = sk.util.WignerD(0, 0)

    theta = np.arange(0, np.pi, 0.01)

    _ = wd.d(theta, 10)


def test_wigner_against_legendre():
    wig = sk.util.WignerD(0, 0)

    theta = np.arange(0, np.pi, 0.01)

    for legidx in range(20):
        true = eval_legendre(legidx, np.cos(theta))

        ours = wig.d(theta, legidx)

        np.testing.assert_array_almost_equal(true, ours)



import numpy as np
import sasktran2 as sk


def test_linear_interpolating_matrix():
    """
    Test that the linear_interpolating_matrix function works as expected by comparing against the
    numpy interp function
    """
    from_grid = np.array([0.0, 1000.0, 3000.0, 5000.0, 10000.0])

    to_grid = np.array([-20.0, 500.0, 3000.0, 4000.0, 8000.0, 12000.0])

    rng = np.random.default_rng()
    vals = rng.random(len(from_grid))

    M = sk.util.interpolation.linear_interpolating_matrix(from_grid, to_grid, "extend")
    np.testing.assert_array_almost_equal(
        M @ vals, np.interp(to_grid, from_grid, vals, left=vals[0], right=vals[-1])
    )

    M = sk.util.interpolation.linear_interpolating_matrix(from_grid, to_grid, "zero")
    np.testing.assert_array_almost_equal(
        M @ vals, np.interp(to_grid, from_grid, vals, left=0, right=0)
    )



import numpy as np

import sasktran2 as sk
from sasktran2.constants import K_BOLTZMANN

from . import database


class O3DBM(database.OpticalDatabaseGenericAbsorber):
    def __init__(self) -> None:
        """
        Tabulated high resolution cross-sections of O3 measured by Daumont, Brion and Malicet in the early 1990's [1].
        The wavelength range slightly varies with temperature but covers the entire UV to NIR region, from 194.50 nm to
        830.00 nm at 0.01 to 0.02 nm resolution. The cross-section data were collected at 0.01-0.02 nm resolution and
        each wavelength/cross-section table varies in size from 22,052 to 63,501 entries. The data consists of 5 tables
        of wavelength versus cross-section for 5 temperatures.

        Notes
        -----
        Temerature Range
            Measurements are provided at 5 temperatures covering typical stratospheric and
            tropospheric conditions:

                | 218 K
                | 228 K
                | 243 K
                | 273 K
                | 295 K

        Wavelength Range
            The wavelength range of each temperature table is slightly different and is given below.
            Note that most of the temperature variation occurs in the Huggins band
            between 315 and 360 nm:

                | 218K -> 194.50nm to 650.01nm
                | 228K -> 194.50nm to 520.01nm
                | 243K -> 194.50nm to 519.01nm
                | 273K -> 299.50nm to 520.01nm
                | 295K -> 195.00nm to 830.00nm

            We looked into temperature interpolation and while DBM suggest that a quadratic interpolation scheme [3] they do
            not indicate an explicit technique. We tested several quadratic fitting routines and found that a truncated linear
            fit in temperature was visually more appealing than any of the quadratic fits and had none of the undesirable
            artifacts (excessive curvature etc.) that naturally arise with quadratic curve fitting. Consequently this object
            uses a truncated linear fit in temperature.

        Data Source
            These data are an exact replication of the data files:

                | O3_CRS_BDM_218K.dat
                | O3_CRS_BDM_228K.dat
                | O3_CRS_BDM_243K.dat
                | O3_CRS_BDM_273K.dat
                | O3_CRS_BDM_295K.dat

            Data is from the IGACO site, http://igaco-o3.fmi.fi/ACSO/cross_sections.html. The files were copied on
            July 16-July 25 2012.

        References
        ----------
        .. [1] Daumont, D., et al. "Ozone UV spectroscopy I: Absorption cross-sections at room temperature."
                Journal of Atmospheric Chemistry 15.2 (1992): 145-155.
        .. [2] Brion, J., et al. "High-resolution laboratory absorption cross section of O3. Temperature effect."
                Chemical physics letters 213.5 - 6(1993): 610-612.
        .. [3] Malicet, J., et al. "Ozone UV spectroscopy. II. Absorption cross-sections and temperature dependence."
                Journal of Atmospheric Chemistry 21.3 (1995): 263-273.
        .. [4] Brion, J., et al. "Absorption spectra measurements for the ozone molecule in the 350-830 nm region."
                Journal of Atmospheric Chemistry 30.2 (1998): 291-299.

        Raises
        ------
        OSError
            If the file could not be found
        """
        dbm_file = sk.appconfig.database_root().joinpath("cross_sections/o3/dbm.nc")

        if dbm_file.exists():
            super().__init__(dbm_file)
        else:
            msg = "Could not find DBM file"
            raise OSError(msg)


class NO2Vandaele(database.OpticalDatabaseGenericAbsorber):
    def __init__(self) -> None:
        """
        Calculates the absorption cross section of NO2 molecules from 230 nm to 1000 nm at 220 K to 294 K following [1]

        .. [1] Vandaele, Ann Carine, et al. "Measurements of the NO2 absorption cross-section from 42 000 cm-1 to
            10 000 cm-1 (238-1000 nm) at 220 K and 294 K." Journal of Quantitative Spectroscopy and Radiative Transfer
            59.3-5 (1998): 171-184.

        Raises
        ------
        OSError
            If the Vandaele file cannot be found
        """
        v_file = sk.appconfig.database_root().joinpath("cross_sections/no2/vandaele.nc")

        if v_file.exists():
            super().__init__(v_file)
        else:
            msg = "Could not find Vandaele file"
            raise OSError(msg)


class HITRANUV(database.OpticalDatabaseGenericAbsorber):
    def __init__(self, name: str, version: str = "2022") -> None:
        """
        HITRAN UV cross-sections for a given constituent.  The constituent must be one of the following:
        ["O3", "NO2", "BrO", "SO2"]

        These are the files that you obtain from the HITRAN website when selecting UV cross sections.
        Note that usually these are not very good, and you should use a specific database for each constituent instead
        such as O3DBM.  The HITRAN UV cross sections are only provided for convenience.

        Specific Issues
            O3
                We have noticed the Ozone cross sections are unreasonably large in the region < 310 nm, and does not match the source data.
                We do not recommend using this.
            NO2
                Even though the HITRAN database specifically says all cross sections are given in vacuum wavenumbers, this one
                seems to be specified in air wavenumbers.  We have converted it to vacuum wavenumbers manually. After doing so
                it matches the Vandaele source files relatively well.

        Parameters
        ----------
        name : str
            Constituent
        version : str, optional
            HITRAN version number, currently only "2022" is supported, by default "2022"

        Raises
        ------
        OSError
            If the databases are not installed
        """
        data_file = sk.appconfig.database_root().joinpath(
            f"cross_sections/{name}/hitran{version}.nc"
        )

        if data_file.exists():
            super().__init__(data_file)
        else:
            msg = f"Could not find HITRAN UV database for {name} at {data_file}"
            raise OSError(msg)


class HITRANTabulated(database.OpticalDatabaseGenericAbsorber):
    def __init__(self, name: str, res="01nm") -> None:
        """
        Loads in a database tabulated from HITRAN line entries as a function of pressure/temperature that have been
        reduced to a given resolution.

        This requires the extended databases to be downloaded, e.g. by running sk.appconfig.download_extended_databases()

        Currently supported species and resolutions are:

            | H2O : 01nm
            | O2 : 01nm

        Parameters
        ----------
        name : str
            Species name
        res : str, optional
            Resolution of the database to load in, by default "01nm"

        Raises
        ------
        OSError
            If the specified file cannot be found
        """
        data_file = sk.appconfig.database_root().joinpath(
            f"cross_sections/{name.lower()}/hitran_{res}_res.nc"
        )

        if data_file.exists():
            super().__init__(data_file)
        else:
            msg = f"Could not find HITRAN database for {name} at {data_file}"
            raise OSError(msg)


def pressure_temperature_to_numberdensity(
    pressure_pa: np.array, temperature_k: np.array, include_derivatives=False
) -> np.array:
    """
    Converts pressure and temperature to number density using the ideal gas law

    Parameters
    ----------
    pressure_pa : np.array
        Pressure in [Pa]
    temperature_k : np.array
        Temperature in [K]
    include_derivatives: bool, optional
        If set to true, the derivative of the number density with respect to pressure and temperature will be returned
        alongside the numberdensity.  The return signature is then np.array, np.array, np.array

    Returns
    -------
    np.array, if include_derivatives is False
        Number density in [molecules / m^3]

    np.array, np.array, np.array if include_derivatives is True
        Number density, dN/dP, dN/dT in [molecules / m^3], [molecules / m^3 / Pa], [molecules / m^3 / K]
    """
    # Ideal gas law is PV = N k T
    # Or (N/V) = P / (k T)

    N = pressure_pa / (K_BOLTZMANN * temperature_k)

    if not include_derivatives:
        return N

    dN_dP = 1 / (K_BOLTZMANN * temperature_k)
    dN_dT = -pressure_pa / (K_BOLTZMANN * temperature_k**2)

    return N, dN_dP, dN_dT


def air_wavelength_to_vacuum_wavelength(wavelength_nm: np.array) -> np.array:
    """
    Converts wavelength specified in Air at STP to wavelengths at vacuum

    Parameters
    ----------
    wavelength_nm : np.array
        Wavelength in air at STP

    Returns
    -------
    np.array
        Vacuum wavelengths
    """
    s = 1e4 / (wavelength_nm * 10)  # Convert to angstroms

    refac_index = (
        1
        + 0.00008336624212083
        + 0.02408926869968 / (130.1065924522 - s**2)
        + 0.0001599740894897 / (38.92568793293 - s**2)
    )

    return wavelength_nm * refac_index


def vacuum_wavelength_to_air_wavelength(wavelength_nm: np.array) -> np.array:
    """
    Converts wavelength specified in vacuum to wavelengths at air at STP

    Parameters
    ----------
    wavelength_nm : np.array
        Wavelengths in vacuum

    Returns
    -------
    np.array
        Wavelengths in air at STP
    """
    s = 1e4 / (wavelength_nm * 10)  # Convert to angstroms

    n = 1 + 0.0000834254 + 0.02406147 / (130 - s**2) + 0.00015998 / (38.9 - s**2)

    return wavelength_nm / n



import numpy as np
import pytest
import sasktran2 as sk


def _test_scenarios():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 4
    config.delta_m_scaling = False

    altitude_grid = np.arange(0, 65001, 1000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    wavel = np.array([310, 330, 350, 600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere.surface.albedo[:] = 0.3

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


@pytest.mark.skip()
def test_scatterer_extinction_altitude_construction():
    """
    Test that the ExtinctionScatterer class can be constructed and basic functionality works
    """
    alts = np.arange(0, 100001, 1000)
    ext_wavel = 525

    ext_profile = sk.climatology.glossac.stratospheric_background(
        5, 20, alts, ext_wavel
    )

    mie = sk.optical.database.OpticalDatabaseGenericScatterer(
        sk.appconfig.database_root().joinpath("cross_sections/mie/sulfate_test.nc")
    )
    radius = np.ones_like(alts) * 105
    const = sk.constituent.ExtinctionScatterer(
        mie, alts, ext_profile, ext_wavel, lognormal_median_radius=radius
    )

    assert len(const.number_density) == len(alts)
    assert len(const.extinction_per_m) == len(alts)


def test_scatterer_extinction_in_engine():
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        _ = engine.calculate_radiance(scen["atmosphere"])


def test_scatterer_extinction_wf_native_grid():
    """
    Tests the derivatives for a scatterer on the native grid
    """

    scens = _test_scenarios()

    for scen in scens:
        altitude_grid = scen["atmosphere"].model_geometry.altitudes()
        atmosphere = scen["atmosphere"]

        atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
            sk.optical.O3DBM(), altitude_grid, np.ones_like(altitude_grid) * 1e-6
        )

        atmosphere["strat_aerosol"] = sk.test_util.scenarios.test_aerosol_constituent(
            altitude_grid, True
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["strat_aerosol"].extinction_per_m,
            0.0001,
            engine,
            atmosphere,
            "wf_strat_aerosol_extinction",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_strat_aerosol_extinction"],
            radiance["wf_strat_aerosol_extinction_numeric"],
            wf_dim="strat_aerosol_altitude",
            decimal=5,
        )


def test_scatterer_extinction_wf_interpolated_grid():
    """
    Tests the derivatives for a scatterer on an interpolated grid
    """

    scens = _test_scenarios()

    new_altitudes = np.array([0, 10000, 30000, 70000])

    for scen in scens:
        altitude_grid = scen["atmosphere"].model_geometry.altitudes()
        atmosphere = scen["atmosphere"]

        atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
            sk.optical.O3DBM(), altitude_grid, np.ones_like(altitude_grid) * 1e-6
        )

        atmosphere["strat_aerosol"] = sk.test_util.scenarios.test_aerosol_constituent(
            new_altitudes, True
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["strat_aerosol"].extinction_per_m,
            0.0001,
            engine,
            atmosphere,
            "wf_strat_aerosol_extinction",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_strat_aerosol_extinction"],
            radiance["wf_strat_aerosol_extinction_numeric"],
            wf_dim="strat_aerosol_altitude",
            decimal=4,
        )


def test_scatterer_numden_wf_interpolated_grid():
    """
    Tests the derivatives for a number density scatterer on an interpolated grid
    """

    scens = _test_scenarios()
    new_altitudes = np.array([0, 10000, 30000, 70000])

    for scen in scens:
        altitude_grid = scen["atmosphere"].model_geometry.altitudes()
        atmosphere = scen["atmosphere"]

        atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
            sk.optical.O3DBM(), altitude_grid, np.ones_like(altitude_grid) * 1e-6
        )

        atmosphere["strat_aerosol"] = sk.test_util.scenarios.test_aerosol_constituent(
            new_altitudes, False
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["strat_aerosol"].number_density,
            0.00001,
            engine,
            atmosphere,
            "wf_strat_aerosol_number_density",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_strat_aerosol_number_density"],
            radiance["wf_strat_aerosol_number_density_numeric"],
            wf_dim="strat_aerosol_altitude",
            decimal=4,
        )



import numpy as np

from sasktran2 import Atmosphere
from sasktran2.atmosphere import (
    InterpolatedDerivativeMapping,
    NativeGridDerivative,
)
from sasktran2.optical import pressure_temperature_to_numberdensity
from sasktran2.optical.base import OpticalProperty
from sasktran2.util.interpolation import linear_interpolating_matrix

from .base import Constituent


class VMRAltitudeAbsorber(Constituent):
    def __init__(
        self,
        optical_property: OpticalProperty,
        altitudes_m: np.array,
        vmr: np.array,
        out_of_bounds_mode: str = "zero",
    ) -> None:
        """
        An atmospheric constituent that is specified through volume mixing ratio (VMR) on an altitude grid.
        The altitude grid need not match the global atmospheric grid, in the case they are different the VMRs
        will be linearly interpolated to the atmospheric grid locations.

        The constituent has an associated OpticalProperty that defines the absorption cross section as a function
        of wavelength.

        Parameters
        ----------
        optical_property : OpticalProperty
            An object that provides absorption cross sections as a function of wavelength
        altitudes_m : np.array
            Altitudes in [m] that the VMR is specified on
        vmr : np.array
            Volume mixing ratio
        out_of_bounds_mode : str, optional
            One of ['zero', 'extend'], 'zero' will set the VMR to be 0 on any altitudes that are
            out of bounds.  'extend' will extend the last or first VMR value.  By default 'zero'
        """
        super().__init__()

        self._out_of_bounds_mode = out_of_bounds_mode
        self._altitudes_m = altitudes_m
        self._vmr = vmr
        self._optical_property = optical_property

    @property
    def vmr(self):
        return self._vmr

    @vmr.setter
    def vmr(self, vmr: np.array):
        self._vmr = vmr

    def add_to_atmosphere(self, atmo: Atmosphere):
        self._optical_quants = self._optical_property.atmosphere_quantities(atmo)

        if atmo.pressure_pa is None or atmo.temperature_k is None:
            msg = "Both pressure_pa and temperature_k have to be specified in the atmosphere to use VMRAltitudeAbsorber"
            raise ValueError(msg)

        number_density = pressure_temperature_to_numberdensity(
            atmo.pressure_pa, atmo.temperature_k
        )

        interp_matrix = linear_interpolating_matrix(
            self._altitudes_m,
            atmo.model_geometry.altitudes(),
            self._out_of_bounds_mode.lower(),
        )

        interp_vmr = interp_matrix @ self._vmr

        atmo.storage.total_extinction += (
            self._optical_quants.extinction
            * (number_density * interp_vmr)[:, np.newaxis]
        )

    def register_derivative(self, atmo: Atmosphere, name: str):
        number_density, dN_dP, dN_dT = pressure_temperature_to_numberdensity(
            atmo.pressure_pa, atmo.temperature_k, include_derivatives=True
        )
        interp_matrix = linear_interpolating_matrix(
            self._altitudes_m,
            atmo.model_geometry.altitudes(),
            self._out_of_bounds_mode.lower(),
        )
        derivs = {}

        derivs["vmr"] = InterpolatedDerivativeMapping(
            NativeGridDerivative(
                d_extinction=self._optical_quants.extinction
                * number_density[:, np.newaxis],
                d_ssa=self._optical_quants.extinction
                * (self._optical_quants.ssa - atmo.storage.ssa)
                / atmo.storage.total_extinction
                * number_density[:, np.newaxis],
            ),
            interpolating_matrix=interp_matrix,
            interp_dim="altitude",
            result_dim=f"{name}_altitude",
        )

        interp_vmr = interp_matrix @ self._vmr

        deriv_names = []
        if atmo.calculate_pressure_derivative:
            deriv_names.append("pressure_pa")
        if atmo.calculate_temperature_derivative:
            deriv_names.append("temperature_k")

        # Contributions from the change in number density due to a constant
        # VMR and changing pressure/temperature
        for deriv_name, vert_factor in zip(deriv_names, [dN_dP, dN_dT]):
            derivs[deriv_name] = InterpolatedDerivativeMapping(
                NativeGridDerivative(
                    d_extinction=self._optical_quants.extinction,
                    d_ssa=self._optical_quants.extinction
                    * (self._optical_quants.ssa - atmo.storage.ssa)
                    / atmo.storage.total_extinction,
                ),
                interpolating_matrix=np.eye(len(number_density))
                * (vert_factor * interp_vmr)[np.newaxis, :],
                interp_dim="altitude",
                result_dim="altitude",
                summable=True,
            )

        if len(deriv_names) > 0:
            optical_derivs = self._optical_property.optical_derivatives(atmo=atmo)

            for key, val in optical_derivs.items():
                # We only get d_extinction from the optical property, have to set d_ssa accordingly
                val.d_ssa = (
                    val.d_extinction
                    * (self._optical_quants.ssa - atmo.storage.ssa)
                    / atmo.storage.total_extinction
                )

                # Assume that all optical derivs are summable, I can't think of any that aren't right now.
                # Give it a unique key to help with the summing
                derivs[f"{key}_xs"] = InterpolatedDerivativeMapping(
                    val,
                    interpolating_matrix=np.eye(len(number_density))
                    * (number_density * interp_vmr)[np.newaxis, :],
                    interp_dim="altitude",
                    result_dim="altitude",
                    summable=True,
                    name=key,
                )

        return derivs



import numpy as np
import sasktran2 as sk


def test_mie_construction():
    mie = sk.mie.LinearizedMie()
    size_param = np.array([3.0, 4.0, 5.0])
    cos_angles = np.linspace(-1, 1, 100)
    refractive_index = 1.5 + 0.0j

    output = mie.calculate(size_param, refractive_index, cos_angles, True)

    np.testing.assert_allclose(output.size_parameter, size_param)
    np.testing.assert_allclose(output.cos_angles, cos_angles)


def test_qext_q_Sca_s1_s2():
    mie = sk.mie.LinearizedMie()
    refractive_index = 0.75
    size_param = np.array(
        [0.101, 0.2]
    )  # not checking values for 0.2, just there to show that we can calculate with multiple size param and angles
    angles = np.arange(0, 181, 30)
    cos_angles = np.cos(angles * np.pi / 180)
    output = mie.calculate(size_param, refractive_index, cos_angles, True)

    np.testing.assert_allclose(output.values.Qext[0], 0.000008, atol=1e-06, rtol=0)
    np.testing.assert_allclose(output.values.Qsca[0], 0.000008, atol=1e-06, rtol=0)

    np.testing.assert_allclose(
        output.values.S1[0][0].real, 2.048754e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][0].imag, -1.756419e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][0].real, 2.048754e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][0].imag, -1.756419e-04, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][1].real, 2.048754e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][1].imag, -1.755965e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][1].real, 1.774273e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][1].imag, -1.520629e-04, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][2].real, 2.048753e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][2].imag, -1.754726e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][2].real, 1.024377e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][2].imag, -8.771198e-05, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][3].real, 2.048751e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][3].imag, -1.753033e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][3].real, 1.845057e-15, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][3].imag, -3.247270e-08, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][4].real, 2.048750e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][4].imag, -1.751341e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][4].real, -1.024375e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][4].imag, 8.759147e-05, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][5].real, 2.048749e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][5].imag, -1.750103e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][5].real, -1.774269e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][5].imag, 1.515715e-04, atol=1e-06, rtol=0
    )

    np.testing.assert_allclose(
        output.values.S1[0][4].real, 2.048749e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S1[0][5].imag, -1.749650e-04, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][6].real, -2.048749e-08, atol=1e-06, rtol=0
    )
    np.testing.assert_allclose(
        output.values.S2[0][6].imag, 1.749650e-04, atol=1e-06, rtol=0
    )



from sasktran2 import Atmosphere
from sasktran2.optical.base import OpticalProperty

from .base import Constituent


class AltitudeAbsorber(Constituent):
    def __init__(self, optical_property: OpticalProperty) -> None:
        super().__init__()

        self._optical_property = optical_property

    def add_to_atmosphere(self, atmo: Atmosphere):
        return super().add_to_atmosphere(atmo)

    def register_derivative(self, atmo: Atmosphere, name: str):  # noqa: ARG002
        return super().register_derivative(atmo)



import numpy as np

from sasktran2 import Atmosphere
from sasktran2.atmosphere import (
    NativeGridDerivative,
    SurfaceDerivativeMapping,
)
from sasktran2.util.interpolation import linear_interpolating_matrix

from .base import Constituent


class LambertianSurface(Constituent):
    def __init__(
        self,
        albedo: np.array,
        wavelengths_nm: np.array = None,
        wavenumbers_cminv: np.array = None,
        out_of_bounds_mode="zero",
    ) -> None:
        """
        A Lambertian surface that is defined by albedo at discrete grid points.

        This can either operate in a "scalar" mode where the albedo is constant in wavelength,
        a "native" mode where the albedo is defined on the same grid as the atmosphere, or
        an "interpolated" mode where the albedo is interpolated either in wavenumber or wavelength

        Parameters
        ----------
        albedo : np.array
            Surface albedo.  Can be a scalar to indicate it is constant in wavelength.  If set to an
            array it must either match the atmosphere wavelength grid dimension, or one of
            wavelengths_nm or wavenumbers_cminv must be set.
        wavelengths_nm : np.array, optional
            Wavelengths in [nm] that the albedo is specified at, by default None
        wavenumbers_cminv : np.array, optional
            Wavenumbers in inverse cm that the albedo is specified at, by default None
        out_of_bounds_mode : str, optional
            One of ["extend" or "zero"], "extend" will extend the last/first value if we are
            interpolating outside the grid. "zero" will set the albedo to 0 outside of the
            grid boundaries, by default "zero"
        """
        super().__init__()

        self._out_of_bounds_mode = out_of_bounds_mode
        self._albedo = np.atleast_1d(albedo)

        if wavelengths_nm is not None:
            self._x = np.atleast_1d(wavelengths_nm)
            self._interp_var = "wavelengths_nm"
        elif wavenumbers_cminv is not None:
            self._x = np.atleast_1d(wavenumbers_cminv)
            self._interp_var = "wavenumbers_cminv"
        else:
            if len(self._albedo) == 1:
                self._interp_var = "constant"
            else:
                self._interp_var = "native"

    @property
    def albedo(self) -> np.array:
        return self._albedo

    @albedo.setter
    def albedo(self, albedo: np.array):
        self._albedo = np.atleast_1d(albedo)

    def _interpolating_matrix(self, atmo: Atmosphere):
        if self._interp_var == "constant":
            # Don't have to interpolate, just assign the constant value
            return np.ones_like(atmo.surface.albedo).reshape(-1, 1)
        if self._interp_var == "native":
            # Also can just assign the user value, but first make sure that it is the correct length
            if len(self._albedo) != len(atmo.surface.albedo):
                msg = "The number of albedo values must match the number of wavelengths in the atmosphere"
                raise ValueError(msg)
            return np.eye(len(self._albedo))
        # Now we have to interpolate
        grid_x = getattr(atmo, self._interp_var)

        if grid_x is None:
            msg = f"The atmosphere does not have {self._interp_var} defined, cannot interpolate the user albedo"
            raise ValueError(msg)
        return linear_interpolating_matrix(self._x, grid_x, self._out_of_bounds_mode)

    def add_to_atmosphere(self, atmo: Atmosphere):
        interp_matrix = self._interpolating_matrix(atmo)

        atmo.surface.albedo[:] = interp_matrix @ self._albedo

    def register_derivative(self, atmo: Atmosphere, name: str):
        # Start by constructing the interpolation matrix
        interp_matrix = self._interpolating_matrix(atmo)

        derivs = {}

        derivs["albedo"] = SurfaceDerivativeMapping(
            NativeGridDerivative(d_albedo=np.ones_like(atmo.surface.albedo)),
            interpolating_matrix=interp_matrix,
            interp_dim="wavelength",
            result_dim=f"{name}_wavelength",
        )

        return derivs



import logging
import urllib.request
import zipfile
from pathlib import Path

import appdirs
import yaml
from packaging import version

APPDIRS = appdirs.AppDirs(appname="sasktran2", appauthor="usask-arg")


def user_config_file_location() -> Path:
    """
    File location of the user config file
    """
    user_config_dir = APPDIRS.user_config_dir

    return Path(user_config_dir).joinpath("config.yml")


def load_user_config():
    """
    Opens the user config file as a dictionary
    """
    user_config_file = user_config_file_location()

    try:
        with Path.open(user_config_file) as f:
            if version.parse(yaml.__version__) > version.parse("5"):
                config = yaml.load(f, Loader=yaml.FullLoader)
            else:
                config = yaml.load(f)

        if config is not None:
            return config
        return {}
    except FileNotFoundError:
        return {}


def save_user_config(user_config: dict):
    """
    Saves the user config from a dictionry
    """
    user_config_file = user_config_file_location()

    Path(user_config_file).parent.mkdir(exist_ok=True, parents=True)

    with Path.open(user_config_file, "w") as f:
        yaml.dump(user_config, f, default_flow_style=False)


def database_root() -> Path:
    dir = load_user_config().get("database_root", None)

    if dir is None:
        return None
    return Path(dir)


def are_extended_db_downloaded() -> bool:
    """
    Checks if the extended databases are downloaded
    """
    if database_root() is None:
        return False
    return database_root().joinpath("cross_sections/h2o/hitran_01nm_res.nc").exists()


def download_standard_databases(version: str = "latest"):
    user_config = load_user_config()

    if "database_root" in user_config:
        data_directory = Path(user_config["database_root"])
    else:
        data_directory = Path(APPDIRS.user_data_dir).joinpath("database")

    if not data_directory.exists():
        data_directory.mkdir(parents=True)

    output_file_temp = data_directory.joinpath(f"v_{version}.zip")

    if output_file_temp.exists():
        output_file_temp.unlink()

    try:
        urllib.request.urlretrieve(
            f"https://arg.usask.ca/sasktranfiles/sasktran2_db/v_{version}.zip",
            filename=output_file_temp.as_posix(),
        )
        with zipfile.ZipFile(output_file_temp.as_posix(), "r") as zip_ref:
            zip_ref.extractall(data_directory)
    except Exception as e:
        logging.exception(e)

    if output_file_temp.exists():
        output_file_temp.unlink()

    user_config["database_root"] = data_directory.as_posix()
    save_user_config(user_config)


def download_extended_databases(version: str = "latest"):
    db_root = database_root()

    if db_root is None:
        msg = "Database root not set"
        raise OSError(msg)

    output_file_temp = db_root.joinpath(f"v_{version}_extended.zip")

    if output_file_temp.exists():
        output_file_temp.unlink()

    try:
        urllib.request.urlretrieve(
            f"https://arg.usask.ca/sasktranfiles/sasktran2_db/v_{version}_extended.zip",
            filename=output_file_temp.as_posix(),
        )
        with zipfile.ZipFile(output_file_temp.as_posix(), "r") as zip_ref:
            zip_ref.extractall(db_root)
    except Exception as e:
        logging.exception(e)

    if output_file_temp.exists():
        output_file_temp.unlink()



import numpy as np
import sasktran2 as sk


def test_amf_basic():
    """
    Verifies that the AMF constituent runs without issue.  We do not check the results here
    """
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates

    altitude_grid = np.arange(0, 65001, 1000.0)
    geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6327000,
        altitude_grid_m=altitude_grid,
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()
    viewing_geo.add_ray(
        sk.GroundViewingSolar(
            cos_sza=0.6,
            relative_azimuth=0,
            cos_viewing_zenith=-0.8,
            observer_altitude_m=200000,
        )
    )

    engine = sk.Engine(config, geometry, viewing_geo)

    wavel = np.array([330])
    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)
    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere["amf"] = sk.constituent.AirMassFactor()
    atmosphere.surface.albedo[:] = 0.3

    _ = engine.calculate_radiance(atmosphere)



import numpy as np
import sasktran2 as sk


def test_atmosphere_construction():
    config = sk.Config()

    altitude_grid = np.arange(0, 100001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )
    _ = sk.Atmosphere(geometry, config, numwavel=17)


def test_atmosphere_vector_construction():
    config = sk.Config()
    config.num_stokes = 3

    altitude_grid = np.arange(0, 100001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )
    atmosphere = sk.Atmosphere(geometry, config, numwavel=17)

    _ = atmosphere.storage



import numpy as np
import sasktran2 as sk


def _wgs84() -> sk.Geodetic:
    return sk.WGS84()


def test_geodetic_construction():
    _ = _wgs84()


def test_geodetic_from_lat_lon_alt():
    geo = _wgs84()

    # Test case from sasktran1 geodetic
    geo.from_lat_lon_alt(latitude=60, longitude=30, altitude=0)

    np.testing.assert_almost_equal(
        geo.location, np.array([2768773.79083189, 1598552.29346197, 5500477.13393864])
    )
    np.testing.assert_almost_equal(geo.local_south, np.array([0.75, 0.4330127, -0.5]))
    np.testing.assert_almost_equal(geo.local_up, np.array([0.4330127, 0.25, 0.8660254]))
    np.testing.assert_almost_equal(
        geo.local_west, np.array([5.00000000e-01, -8.66025404e-01, -2.77555756e-17])
    )

    # Check internal consistency
    lats = [-80, 0, 80]
    lons = [0, 180, 330]
    alts = [0, 10000, 80000]

    for lat, lon, alt in zip(lats, lons, alts):
        geo.from_lat_lon_alt(latitude=lat, longitude=lon, altitude=alt)

        np.testing.assert_almost_equal(geo.latitude, lat)
        np.testing.assert_almost_equal(geo.longitude, lon)
        np.testing.assert_almost_equal(geo.altitude, alt)


def test_geodetic_from_xyz():
    geo = _wgs84()

    geo.from_xyz([5797230.47518212, -2110019.3341472, -1642001.16317228])

    np.testing.assert_almost_equal(geo.latitude, -14.999999973747736)
    np.testing.assert_almost_equal(geo.longitude, 340.00000000000006)
    np.testing.assert_almost_equal(geo.altitude, 7344.999610390202)


def test_geodetic_from_tangent_altitude():
    geo = _wgs84()

    look = geo.from_tangent_altitude(
        15322,
        [3.676013154788849600e005, 1.009976313640051500e006, -6.871601202127538600e006],
        [0, 0, 1],
    )

    np.testing.assert_almost_equal(look, np.array([0.28880556, 0.79348676, 0.53569591]))
    np.testing.assert_almost_equal(
        geo.location, np.array([1174200.8589215, 3226090.34579105, -5375466.33023017])
    )


def test_geodetic_from_tangent_point():
    geo = _wgs84()

    geo.from_tangent_point(
        [3.676013154788849600e005, 1.009976313640051500e006, -6.871601202127538600e006],
        [
            2.884568631765662100e-001,
            7.925287180643269000e-001,
            5.372996083468238900e-001,
        ],
    )

    np.testing.assert_almost_equal(
        geo.location, np.array([1176731.78951674, 3233044.02045169, -5364459.08730841])
    )



from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import sasktran2 as sk

V_DIR = "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/input/cross_sections/no2/vandaele/"
OUT_DIR = "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/dist/cross_sections/no2/"


all_wv = []
all_xs = []
all_T = []
for no2_file in Path(V_DIR).iterdir():
    data = pd.read_csv(no2_file.as_posix(), header=0, sep="\s+")

    wavelength_nm = 1e7 / data.values[:, 1].astype("float")
    xs_cm2 = data.values[:, 2].astype("float")

    sort_idx = np.argsort(wavelength_nm)

    if "c" in no2_file.stem:
        temp_k = 220
    else:
        temp_k = 294

    all_wv.append(wavelength_nm[sort_idx])
    all_xs.append(xs_cm2[sort_idx])
    all_T.append(temp_k)

combined_wv = np.sort(np.unique(np.hstack(all_wv)))


idx_min_T = np.argmin(all_T)
idx_max_T = np.argmax(all_T)

all_T.append(0)
all_xs.append(all_xs[idx_min_T])
all_wv.append(all_wv[idx_min_T])
all_T.append(1000)
all_xs.append(all_xs[idx_max_T])
all_wv.append(all_wv[idx_max_T])

T_sort = np.argsort(all_T)

combined_xs = np.zeros((len(all_T), len(combined_wv)))


for i in range(len(all_T)):
    combined_xs[i] = np.interp(
        combined_wv, all_wv[i], all_xs[i], left=np.nan, right=np.nan
    )

all_T = np.array(all_T)[T_sort]
combined_xs = combined_xs[T_sort]

# now we have to go through and remove all of the NaN values by extending the other temperature measurements
for i in range(len(combined_wv)):
    isna = np.isnan(combined_xs[:, i])
    if sum(isna) > 0:
        combined_xs[isna, i] = np.interp(
            all_T[isna],
            all_T[~isna],
            combined_xs[~isna, i],
            left=combined_xs[~isna, i][0],
            right=combined_xs[~isna, i][-1],
        )

ds = xr.Dataset(
    {"xs": (["temperature", "wavelength_nm"], combined_xs / 1e4)},
    coords={"temperature": all_T, "wavelength_nm": combined_wv},
)

ds.to_netcdf(Path(OUT_DIR).joinpath("vandaele.nc").as_posix())



# ruff: noqa: F401

from .amf import AirMassFactor
from .lambertiansurface import LambertianSurface
from .numdenscatterer import ExtinctionScatterer, NumberDensityScatterer
from .rayleigh import Rayleigh
from .vmraltitudeabsorber import VMRAltitudeAbsorber



from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import sasktran2 as sk

DBM_DIR = "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/input/cross_sections/o3/dbm/"
OUT_DIR = "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/dist/cross_sections/o3/"


all_wv = []
all_xs = []
all_T = []
for dbm_file in Path(DBM_DIR).iterdir():
    data = pd.read_csv(
        dbm_file,
        header=3,
        converters={
            0: lambda x: float(x.replace("{", "").replace("}", "").replace(" ", ""))
            if x is not None
            else 0,
            1: lambda x: float(x.replace("{", "").replace("}", "").replace(" ", ""))
            if x is not None
            else 0,
        },
        skipfooter=1,
    )

    wavelength_nm = data.values[:, 0].astype("float")
    xs_cm2 = data.values[:, 1].astype("float")

    temp_k = float(dbm_file.stem[7:10])

    all_wv.append(wavelength_nm)
    all_xs.append(xs_cm2)
    all_T.append(temp_k)

combined_wv = np.sort(np.unique(np.hstack(all_wv)))


idx_min_T = np.argmin(all_T)
idx_max_T = np.argmax(all_T)

all_T.append(0)
all_xs.append(all_xs[idx_min_T])
all_wv.append(all_wv[idx_min_T])
all_T.append(1000)
all_xs.append(all_xs[idx_max_T])
all_wv.append(all_wv[idx_max_T])

T_sort = np.argsort(all_T)

combined_xs = np.zeros((len(all_T), len(combined_wv)))


for i in range(len(all_T)):
    combined_xs[i] = np.interp(
        combined_wv, all_wv[i], all_xs[i], left=np.nan, right=np.nan
    )

all_T = np.array(all_T)[T_sort]
combined_xs = combined_xs[T_sort]

# now we have to go through and remove all of the NaN values by extending the other temperature measurements
for i in range(len(combined_wv)):
    isna = np.isnan(combined_xs[:, i])
    if sum(isna) > 0:
        combined_xs[isna, i] = np.interp(
            all_T[isna],
            all_T[~isna],
            combined_xs[~isna, i],
            left=combined_xs[~isna, i][0],
            right=combined_xs[~isna, i][-1],
        )

ds = xr.Dataset(
    {"xs": (["temperature", "wavelength_nm"], combined_xs / 1e4)},
    coords={
        "temperature": all_T,
        "wavelength_nm": sk.optical.air_wavelength_to_vacuum_wavelength(combined_wv),
    },
)

ds.to_netcdf(Path(OUT_DIR).joinpath("dbm.nc").as_posix())



import numpy as np
import pytest
import sasktran2 as sk


@pytest.mark.skipif(
    not sk.appconfig.are_extended_db_downloaded(),
    reason="Extended database not downloaded",
)
def test_hitran_moderate():
    opt_prop = sk.optical.HITRANTabulated("H2O")

    config = sk.Config()

    altitude_grid = np.arange(0, 65001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    # Old SASKTRAN wavelengths are messed up
    wavel = np.arange(280, 800, 0.1)

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    _ = opt_prop.atmosphere_quantities(atmo=atmosphere)



import numpy as np
import pytest
import sasktran2 as sk


@pytest.mark.skip()
def test_glossac_loading():
    """
    Verifies that the glossac data can be loaded in correctly
    """
    _ = sk.climatology.glossac.load_glossac_raw_data()


@pytest.mark.skip()
def test_stratospheric_background():
    """
    Checks the stratospheric background pulling for a variety of conditions
    """
    _ = sk.climatology.glossac.stratospheric_background(
        1, 0.0, np.arange(0, 65000, 1000.0), 525.0
    )



import abc
from typing import Union

import numpy as np
import xarray as xr

import sasktran2 as sk


class Output(abc.ABC):
    """
    An abstract class which defines how output from the radiative transfer Engine is handled.

    The Engine first calls :py:meth:`~internal_output` to get a `pybind11` object which can be passed
    to the internal radiative transfer model. After the calculation is completed, :py:meth:`~post_process`
    is called to get the exact output container that is passed back to the user.

    """

    @abc.abstractmethod
    def internal_output(self):
        pass

    @abc.abstractmethod
    def post_process(
        self,
        atmo: sk.Atmosphere,
        geometry: sk.Geometry1D,
        viewing_geo: sk.ViewingGeometry,
    ):
        pass


class OutputIdeal(Output):
    def __init__(self, nstokes: int):
        """
        The default output container used by SASKTRAN2.  Here radiances (and derivatives) are returned
        back densely for every input line of sight and wavelength without modification.

        Parameters
        ----------
        nstokes : int
            Number of stokes parameters to store for the output quantities

        Raises
        ------
        ValueError
            If nstokes is not 1 or 3
        """
        self._nstokes = nstokes

        if nstokes == 1:
            self._output = sk.OutputIdealStokes_1()
        elif nstokes == 3:
            self._output = sk.OutputIdealStokes_3()
        else:
            msg = "nstokes must be 1 or 3"
            raise ValueError(msg)

    def internal_output(self) -> Union[sk.OutputIdealStokes_1, sk.OutputIdealStokes_3]:
        """
        The internal output object that can be passed to the internal engine

        Returns
        -------
        Union[sk.OutputIdealStokes_1, sk.OutputIdealStokes_3]
        """
        return self._output

    def post_process(
        self,
        atmo: sk.Atmosphere,
        geometry: sk.Geometry1D,
        viewing_geo: sk.ViewingGeometry,
    ):
        """
        Converts the raw output values to a more usable format.  Also performs the mapping of raw model
        derivatives to derivatives of the constituent input quantities.

        Parameters
        ----------
        atmo : sk.Atmosphere
            Atmosphere object after calculation has been performed
        geometry : sk.Geometry1D
            Geometry object
        viewing_geo : sk.ViewingGeometry
            Viewing geometry

        Returns
        -------
        xr.Dataset
        """
        # TODO: A lot of this reshaping and organizing should be done inside C++, not here
        # Maybe some of this should go into the derivative mapping classes too...
        # Have to be a little careful since this can be a slow part of the code

        # Reshape and organize output
        result = xr.Dataset()
        if atmo.wavelengths_nm is not None:
            result.coords["wavelength"] = atmo.wavelengths_nm

        result.coords["stokes"] = ["I", "Q", "U", "V"][: atmo.nstokes]
        result.coords["altitude"] = geometry.altitudes()

        radiance = self._output.radiance.reshape(
            -1, len(viewing_geo.observer_rays), atmo.nstokes
        )
        result["radiance"] = (["wavelength", "los", "stokes"], radiance)

        natmo_grid = atmo.storage.total_extinction.shape[0]
        nwavel = radiance.shape[0]

        d_radiance_raw = self._output.d_radiance.reshape(
            nwavel, len(viewing_geo.observer_rays), atmo.nstokes, -1
        )
        d_radiance_k = d_radiance_raw[:, :, :, :natmo_grid]
        d_radiance_ssa = d_radiance_raw[:, :, :, natmo_grid : 2 * natmo_grid]
        d_radiance_albedo = d_radiance_raw[:, :, :, -1:]

        dk_scaled_by_dk = 1 - atmo.storage.f * atmo.unscaled_ssa
        dssa_scaled_by_dssa = (
            1 - atmo.storage.f
        ) / dk_scaled_by_dk + atmo.storage.f * atmo.storage.ssa / dk_scaled_by_dk

        for constituent_name, deriv_mapping in atmo.deriv_mappings.items():
            if deriv_mapping is None:
                continue

            for deriv_name_raw, mapping in deriv_mapping.items():
                if mapping.name() is not None:
                    deriv_name = mapping.name()
                else:
                    deriv_name = deriv_name_raw

                if mapping.summable:
                    name_to_place_result = mapping.name_prefix + deriv_name
                else:
                    name_to_place_result = (
                        mapping.name_prefix + constituent_name + "_" + deriv_name
                    )

                # Start by calculating the derivative with respect to the quantity on the native grid
                if mapping.is_surface_derivative:
                    np_deriv = (
                        d_radiance_albedo
                        * mapping.native_grid_mapping.d_albedo[
                            :, np.newaxis, np.newaxis, np.newaxis
                        ]
                    )
                else:
                    np_deriv = (
                        mapping.native_grid_mapping.d_extinction * dk_scaled_by_dk
                        - atmo.storage.f
                        * mapping.native_grid_mapping.d_ssa
                        * atmo.unscaled_extinction
                    ).T[:, np.newaxis, np.newaxis, :] * d_radiance_k + (
                        mapping.native_grid_mapping.d_ssa * dssa_scaled_by_dssa
                    ).T[
                        :, np.newaxis, np.newaxis, :
                    ] * d_radiance_ssa

                    if mapping.native_grid_mapping.scat_deriv_index is not None:
                        d_radiance_scat = d_radiance_raw[
                            :,
                            :,
                            :,
                            (2 + mapping.native_grid_mapping.scat_deriv_index)
                            * natmo_grid : (
                                3 + mapping.native_grid_mapping.scat_deriv_index
                            )
                            * natmo_grid,
                        ]

                        np_deriv += (
                            d_radiance_scat
                            * mapping.native_grid_mapping.scat_factor.T[
                                :, np.newaxis, np.newaxis, :
                            ]
                        )

                        if atmo.applied_delta_m_order:
                            # Change in f for the scatterer
                            d_f = (
                                atmo.storage.d_f[
                                    :, :, mapping.native_grid_mapping.scat_deriv_index
                                ]
                                * mapping.native_grid_mapping.scat_factor
                            )

                            # Change in k* due to a change in f
                            np_deriv -= (
                                d_f * atmo.unscaled_ssa * atmo.unscaled_extinction
                            ).T[:, np.newaxis, np.newaxis, :] * d_radiance_k

                            # Change in ssa* due to a change in f
                            np_deriv += (
                                d_f
                                * atmo.unscaled_ssa
                                / (1 - atmo.unscaled_ssa * atmo.storage.f)
                                * (atmo.storage.ssa - 1)
                            ).T[:, np.newaxis, np.newaxis, :] * d_radiance_ssa

                if mapping.log_radiance_space:
                    np_deriv /= radiance[:, :, :, np.newaxis]

                if mapping.is_surface_derivative:
                    mapped_derivative = mapping.map_derivative(
                        np_deriv[:, :, :, 0], ["wavelength", "los", "stokes"]
                    )
                else:
                    mapped_derivative = mapping.map_derivative(
                        np_deriv, ["wavelength", "los", "stokes", "altitude"]
                    )

                if name_to_place_result in result:
                    result[name_to_place_result] += mapped_derivative
                else:
                    result[name_to_place_result] = mapped_derivative

        return result



import numpy as np
import xarray as xr

import sasktran2 as sk


def load_glossac_raw_data(version: str = "2.21") -> xr.Dataset:
    """
    Load's in the GloSSAC raw data for a specific version.  First we check if the file exists within the database directory,
    if it doesn't we load it in through opendap.

    Parameters
    ----------
    version : str, optional
        GloSSAC File version to look for, by default "2.21"

    Returns
    -------
    xr.Dataset
        GloSSAC Dataset
    """
    db_root = sk.appconfig.database_root()

    if db_root is None:
        return xr.open_dataset(
            f"https://opendap.larc.nasa.gov/opendap/GloSSAC/GloSSAC_{version}/GloSSAC_V{version}.nc"
        )

    glossac_file = sk.appconfig.database_root().joinpath(
        f"climatology/glossac/GloSSAC_V{version}.nc"
    )

    if glossac_file.exists():
        return xr.open_dataset(glossac_file)

    return xr.open_dataset(
        f"https://opendap.larc.nasa.gov/opendap/GloSSAC/GloSSAC_{version}/GloSSAC_V{version}.nc"
    )


def stratospheric_background(
    month: int, lat: float, alts: np.array, wavelength_nm: float, version: str = "2.21"
):
    ds = load_glossac_raw_data(version)

    wavel_idx = np.argmin(np.abs(ds["wavelengths_glossac"].to_numpy() - wavelength_nm))
    lat_idx = np.argmin(np.abs(ds["lat"].to_numpy() - lat))

    if np.abs(ds["wavelengths_glossac"].to_numpy()[wavel_idx] - wavelength_nm) > 1e-4:
        msg = f"Could not find wavelength {wavelength_nm} in GloSSAC data.  Valid wavelengths are {ds['wavelengths_glossac'].to_numpy()}"
        raise ValueError(msg)

    data = ds.isel(wavelengths_glossac=wavel_idx)

    background = (
        data["Stratospheric_Background"]
        .sel(month=month)
        .isel(lat=lat_idx)
        .interp(alt=alts / 1000, method="linear")
        .to_numpy()
    )

    background[np.isnan(background)] = 0

    return background / 1e3  # Convert to /m



import numpy as np

from sasktran2 import Atmosphere
from sasktran2.atmosphere import DerivativeMapping, NativeGridDerivative

from .base import Constituent


class AirMassFactor(Constituent):
    def __init__(
        self,
    ) -> None:
        """
        A dummy atmospheric constituent that does not add any terms to the atmosphere, but rather
        enables the calculation of Air Mass Factor derivatives.
        """
        super().__init__()

    def add_to_atmosphere(self, atmo: Atmosphere):
        pass

    def register_derivative(self, atmo: Atmosphere, name: str):  # noqa: ARG002
        altitudes = atmo.model_geometry.altitudes()

        alt_factors = -1 / np.gradient(altitudes)

        # Need to adjust bottom of atmosphere and top to match AMF definition
        alt_factors[0] *= 2
        alt_factors[-1] *= 2

        derivs = {}

        derivs["air_mass_factor"] = DerivativeMapping(
            NativeGridDerivative(
                d_extinction=alt_factors[:, np.newaxis],
                d_ssa=alt_factors[:, np.newaxis]
                * (0 - atmo.storage.ssa)
                / atmo.storage.total_extinction,
            ),
            summable=True,
            log_radiance_space=True,
            name_prefix="",
        )

        return derivs



import glob
import hashlib
import os
import platform
import shutil
import sys
import sysconfig
import tarfile
import textwrap
import zipfile
from tempfile import gettempdir, mkstemp
from urllib.error import HTTPError
from urllib.request import Request, urlopen

OPENBLAS_V = "0.3.23.dev"
OPENBLAS_LONG = "v0.3.23-293-gc2f4bdbb"
BASE_LOC = "https://anaconda.org/scientific-python-nightly-wheels/openblas-libs"
SUPPORTED_PLATFORMS = [
    "linux-aarch64",
    "linux-x86_64",
    "musllinux-x86_64",
    "linux-i686",
    "linux-ppc64le",
    "linux-s390x",
    "win-amd64",
    "win-32",
    "macosx-x86_64",
    "macosx-arm64",
]
IS_32BIT = sys.maxsize < 2**32


def get_plat():
    plat = sysconfig.get_platform()
    plat_split = plat.split("-")
    arch = plat_split[-1]
    if arch == "win32":
        plat = "win-32"
    elif arch in ["universal2", "intel"]:
        plat = f"macosx-{platform.uname().machine}"
    elif len(plat_split) > 2:
        plat = f"{plat_split[0]}-{arch}"
    assert plat in SUPPORTED_PLATFORMS, f"invalid platform {plat}"
    return plat


def get_ilp64():
    if os.environ.get("NPY_USE_BLAS_ILP64", "0") == "0":
        return None
    if IS_32BIT:
        msg = "NPY_USE_BLAS_ILP64 set on 32-bit arch"
        raise RuntimeError(msg)
    return "64_"


def get_manylinux(arch):
    default = "2014"
    ml_ver = os.environ.get("MB_ML_VER", default)
    # XXX For PEP 600 this can be a glibc version
    assert ml_ver in ("2010", "2014", "_2_24"), f"invalid MB_ML_VER {ml_ver}"
    return f"manylinux{ml_ver}_{arch}.tar.gz"


def get_musllinux(arch):
    musl_ver = "1_1"
    return f"musllinux_{musl_ver}_{arch}.tar.gz"


def get_linux(arch):
    # best way of figuring out whether manylinux or musllinux is to look
    # at the packaging tags. If packaging isn't installed (it's not by default)
    # fallback to sysconfig (which may be flakier)
    try:
        from packaging.tags import sys_tags

        tags = list(sys_tags())
        plat = tags[0].platform
    except ImportError:
        # fallback to sysconfig for figuring out if you're using musl
        plat = "manylinux"
        # value could be None
        v = sysconfig.get_config_var("HOST_GNU_TYPE") or ""
        if "musl" in v:
            plat = "musllinux"

    if "manylinux" in plat:
        return get_manylinux(arch)
    elif "musllinux" in plat:
        return get_musllinux(arch)
    return None


def download_openblas(target, plat, ilp64, *, nightly=False):
    osname, arch = plat.split("-")
    fnsuffix = {None: "", "64_": "64_"}[ilp64]
    filename = ""
    headers = {
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 ; "
            "(KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.3"
        )
    }
    suffix = None
    if osname == "linux":
        suffix = get_linux(arch)
        typ = "tar.gz"
    elif osname == "musllinux":
        suffix = get_musllinux(arch)
        typ = "tar.gz"
    elif plat == "macosx-x86_64":
        suffix = "macosx_10_9_x86_64-gf_c469a42.tar.gz"
        typ = "tar.gz"
    elif plat == "macosx-arm64":
        suffix = "macosx_11_0_arm64-gf_5272328.tar.gz"
        typ = "tar.gz"
    elif osname == "win":
        if plat == "win-32":
            suffix = "win32-gcc_8_3_0.zip"
        else:
            suffix = "win_amd64-gcc_10_3_0.zip"
        typ = "zip"

    if not suffix:
        return None
    openblas_version = "HEAD" if nightly else OPENBLAS_LONG
    filename = (
        f"{BASE_LOC}/{openblas_version}/download/"
        f"openblas{fnsuffix}-{openblas_version}-{suffix}"
    )
    print(f"Attempting to download {filename}", file=sys.stderr)
    req = Request(url=filename, headers=headers)
    try:
        response = urlopen(req)
    except HTTPError:
        print(f'Could not download "{filename}"', file=sys.stderr)
        raise
    length = response.getheader("content-length")
    if response.status != 200:
        print(f'Could not download "{filename}"', file=sys.stderr)
        return None
    # print(f"Downloading {length} from {filename}", file=sys.stderr)
    data = response.read()
    # Verify hash
    key = os.path.basename(filename)
    # print("Saving to file", file=sys.stderr)
    with open(target, "wb") as fid:
        fid.write(data)
    return typ


def setup_openblas(plat=get_plat(), ilp64=get_ilp64(), nightly=False):
    """
    Download and setup an openblas library for building. If successful,
    the configuration script will find it automatically.

    Returns
    -------
    msg : str
        path to extracted files on success, otherwise indicates what went wrong
        To determine success, do ``os.path.exists(msg)``
    """
    _, tmp = mkstemp()
    if not plat:
        msg = "unknown platform"
        raise ValueError(msg)
    typ = download_openblas(tmp, plat, ilp64, nightly=nightly)
    if not typ:
        return ""
    osname, arch = plat.split("-")
    if osname == "win":
        if typ != "zip":
            return f"expecting to download zipfile on windows, not {typ}"
        return unpack_windows_zip(tmp, plat)
    else:
        if typ != "tar.gz":
            return "expecting to download tar.gz, not %s" % str(typ)
        return unpack_targz(tmp)


def unpack_windows_zip(fname, plat):
    unzip_base = os.path.join(gettempdir(), "openblas")
    if not os.path.exists(unzip_base):
        os.mkdir(unzip_base)
    with zipfile.ZipFile(fname, "r") as zf:
        zf.extractall(unzip_base)
    if plat == "win-32":
        target = os.path.join(unzip_base, "32")
    else:
        target = os.path.join(unzip_base, "64")
    # Copy the lib to openblas.lib. Once we can properly use pkg-config
    # this will not be needed
    lib = glob.glob(os.path.join(target, "lib", "*.lib"))
    if len(lib) == 1:
        # The 64-bit tarball already has these copied, no need to do it
        for f in lib:
            shutil.copy(f, os.path.join(target, "lib", "openblas.lib"))
            shutil.copy(f, os.path.join(target, "lib", "openblas64_.lib"))
    # Copy the dll from bin to lib so system_info can pick it up
    dll = glob.glob(os.path.join(target, "bin", "*.dll"))
    for f in dll:
        shutil.copy(f, os.path.join(target, "lib"))
    return target


def unpack_targz(fname):
    target = os.path.join(gettempdir(), "openblas")
    if not os.path.exists(target):
        os.mkdir(target)
    with tarfile.open(fname, "r") as zf:
        # Strip common prefix from paths when unpacking
        prefix = os.path.commonpath(zf.getnames())
        extract_tarfile_to(zf, target, prefix)
        return target


def extract_tarfile_to(tarfileobj, target_path, archive_path):
    """Extract TarFile contents under archive_path/ to target_path/"""

    target_path = os.path.abspath(target_path)

    def get_members():
        for member in tarfileobj.getmembers():
            if archive_path:
                norm_path = os.path.normpath(member.name)
                if norm_path.startswith(archive_path + os.path.sep):
                    member.name = norm_path[len(archive_path) + 1 :]
                else:
                    continue

            dst_path = os.path.abspath(os.path.join(target_path, member.name))
            if os.path.commonpath([target_path, dst_path]) != target_path:
                # Path not under target_path, probably contains ../
                continue

            yield member

    tarfileobj.extractall(target_path, members=get_members())


def make_init(dirname):
    """
    Create a _distributor_init.py file for OpenBlas

    Obsoleted by the use of delvewheel in wheel building, which
    adds an equivalent snippet to numpy/__init__.py, but still useful in CI
    """
    with open(os.path.join(dirname, "_distributor_init.py"), "w") as fid:
        fid.write(
            textwrap.dedent(
                """
            '''
            Helper to preload windows dlls to prevent dll not found errors.
            Once a DLL is preloaded, its namespace is made available to any
            subsequent DLL. This file originated in the numpy-wheels repo,
            and is created as part of the scripts that build the wheel.

            '''
            import os
            import glob
            if os.name == 'nt':
                # load any DLL from numpy/../numpy.libs/, if present
                try:
                    from ctypes import WinDLL
                except:
                    pass
                else:
                    basedir = os.path.dirname(__file__)
                    libs_dir = os.path.join(basedir, os.pardir, 'numpy.libs')
                    libs_dir = os.path.abspath(libs_dir)
                    DLL_filenames = []
                    if os.path.isdir(libs_dir):
                        for filename in glob.glob(os.path.join(libs_dir,
                                                               '*openblas*dll')):
                            # NOTE: would it change behavior to load ALL
                            # DLLs at this path vs. the name restriction?
                            WinDLL(os.path.abspath(filename))
                            DLL_filenames.append(filename)
                    if len(DLL_filenames) > 1:
                        import warnings
                        warnings.warn("loaded more than 1 DLL from .libs:"
                                      "\\n%s" % "\\n".join(DLL_filenames),
                                      stacklevel=1)
    """
            )
        )


def test_setup(plats):
    """
    Make sure all the downloadable files needed for wheel building
    exist and can be opened
    """

    def items():
        """yields all combinations of arch, ilp64"""
        for plat in plats:
            yield plat, None
            osname, arch = plat.split("-")
            if arch not in ("i686", "32"):
                yield plat, "64_"

    errs = []
    for plat, ilp64 in items():
        osname, _ = plat.split("-")
        if plat not in plats:
            continue
        target = None
        try:
            try:
                target = setup_openblas(plat, ilp64)
            except Exception as e:
                print(f"Could not setup {plat} with ilp64 {ilp64}, ")
                print(e)
                errs.append(e)
                continue
            if not target:
                msg = f"Could not setup {plat}"
                raise RuntimeError(msg)
            print("success with", plat, ilp64)
            files = glob.glob(os.path.join(target, "lib", "*.a"))
            if not files:
                msg = "No lib/*.a unpacked!"
                raise RuntimeError(msg)
        finally:
            if target:
                if os.path.isfile(target):
                    os.unlink(target)
                else:
                    shutil.rmtree(target)
    if errs:
        raise errs[0]


def test_version(expected_version=None):
    """
    Assert that expected OpenBLAS version is
    actually available via NumPy. Requires threadpoolctl
    """
    import numpy as np
    import threadpoolctl

    data = threadpoolctl.threadpool_info()
    if len(data) != 1:
        if platform.python_implementation() == "PyPy":
            print("Not using OpenBLAS for PyPy in Azure CI, so skip this")
            return
        msg = f"expected single threadpool_info result, got {data}"
        raise ValueError(msg)
    if not expected_version:
        expected_version = OPENBLAS_V
    if data[0]["version"] != expected_version:
        msg = f"expected OpenBLAS version {expected_version}, got {data}"
        raise ValueError(msg)
    print("OK")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Download and expand an OpenBLAS archive for this architecture"
    )
    parser.add_argument(
        "--test",
        nargs="*",
        default=None,
        help='Test different architectures. "all", or any of ' f"{SUPPORTED_PLATFORMS}",
    )
    parser.add_argument(
        "--check_version",
        nargs="?",
        default="",
        help="Check provided OpenBLAS version string against available OpenBLAS",
    )
    parser.add_argument(
        "--nightly", action="store_true", help="If set, use nightly OpenBLAS build."
    )
    args = parser.parse_args()
    if args.check_version != "":
        test_version(args.check_version)
    elif args.test is None:
        print(setup_openblas(nightly=args.nightly))
    else:
        if len(args.test) == 0 or "all" in args.test:
            test_setup(SUPPORTED_PLATFORMS)
        else:
            test_setup(args.test)



import pandas as pd
from pathlib import Path
import numpy as np
import xarray as xr
import sasktran2 as sk

HITRAN_FOLDER = Path(
    "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/input/hitran_uv/"
)
OUT_DIR = Path(
    "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/dist/cross_sections/"
)


for folder in HITRAN_FOLDER.iterdir():
    name = folder.stem

    all_wv = []
    all_xs = []
    all_T = []
    for file in folder.glob("*.xsc"):
        header = pd.read_csv(file, header=None, nrows=1, sep="\s+")

        wv_num_low = header.values[0, 1]
        wv_num_high = header.values[0, 2]

        npoints = header.values[0, 3]
        temperature = header.values[0, 4]

        wv_num = 1e7 / np.linspace(wv_num_low, wv_num_high, npoints)

        if name == "no2":
            wv_num = sk.optical.air_wavelength_to_vacuum_wavelength(wv_num)

        xs_cm2 = pd.read_csv(file, header=None, skiprows=1, sep="\s+").values.flatten()[
            :npoints
        ]

        sort_idx = np.argsort(wv_num)

        all_wv.append(wv_num[sort_idx])
        all_xs.append(xs_cm2[sort_idx])
        all_T.append(temperature)

    if len(all_wv) == 0:
        continue

    combined_wv = np.sort(np.unique(np.hstack(all_wv)))

    idx_min_T = np.argmin(all_T)
    idx_max_T = np.argmax(all_T)

    all_T.append(0)
    all_xs.append(all_xs[idx_min_T])
    all_wv.append(all_wv[idx_min_T])
    all_T.append(1000)
    all_xs.append(all_xs[idx_max_T])
    all_wv.append(all_wv[idx_max_T])

    T_sort = np.argsort(all_T)

    combined_xs = np.zeros((len(all_T), len(combined_wv)))

    for i in range(len(all_T)):
        combined_xs[i] = np.interp(
            combined_wv, all_wv[i], all_xs[i], left=np.nan, right=np.nan
        )

    all_T = np.array(all_T)[T_sort]
    combined_xs = combined_xs[T_sort]

    # now we have to go through and remove all of the NaN values by extending the other temperature measurements
    for i in range(len(combined_wv)):
        isna = np.isnan(combined_xs[:, i])
        if sum(isna) > 0:
            combined_xs[isna, i] = np.interp(
                all_T[isna],
                all_T[~isna],
                combined_xs[~isna, i],
                left=combined_xs[~isna, i][0],
                right=combined_xs[~isna, i][-1],
            )

    ds = xr.Dataset(
        {"xs": (["temperature", "wavelength_nm"], combined_xs / 1e4)},
        coords={"temperature": all_T, "wavelength_nm": combined_wv},
    )

    species_dir = OUT_DIR.joinpath(name)

    if not species_dir.exists():
        species_dir.mkdir()

    ds.to_netcdf(species_dir.joinpath("hitran2022.nc").as_posix())



import numpy as np
import sasktran2 as sk
from sasktran2.test_util.wf import validate_wf


def _raw_scenarios() -> list:
    """
    Defines the set of scenarios that we later on test dI/dk, dI/dssa, dI/dleg_coeff, dI/dalbedo on

    Right now we test for

    Configurations

    - Single scatter only
    - DO Source

    Viewing geometries

    - Limb viewing
    - Nadir viewing

    Model Geometry

    - 1D spherical

    Atmosphere

    - ficticious atmosphere with only Rayleigh scattering

    """
    configs = []

    configs.append(sk.Config())
    configs[-1].multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    configs[-1].num_streams = 4
    configs[-1].single_scatter_source = sk.SingleScatterSource.NoSource
    configs[-1].num_stokes = 3

    configs.append(sk.Config())
    configs[-1].multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    configs[-1].num_streams = 4

    configs.append(sk.Config())

    for config in configs:
        config.num_singlescatter_moments = 4

    geometrys = []
    altitude_grid = np.arange(0, 65001, 5000.0)

    geometrys.append(
        sk.Geometry1D(
            0.6,
            0,
            6327000,
            altitude_grid,
            sk.InterpolationMethod.LinearInterpolation,
            sk.GeometryType.Spherical,
        )
    )

    viewing_geos = []
    viewing_geos.append(sk.ViewingGeometry())

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geos[-1].add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    viewing_geos.append(sk.ViewingGeometry())
    viewing_geos[-1].add_ray(sk.GroundViewingSolar(0.6, 0, 0.8, 200000))

    scen = []

    for config in configs:
        for geometry in geometrys:
            for viewing_geo in viewing_geos:
                scen.append(
                    {
                        "config": config,
                        "geometry": geometry,
                        "viewing_geo": viewing_geo,
                        "atmosphere": sk.test_util.scenarios.default_pure_scattering_atmosphere(
                            config, geometry, 0.8
                        ),
                    }
                )

    # Add an additional scenario that is for plane parallel geometry and the DO source
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.single_scatter_source = sk.SingleScatterSource.DiscreteOrdinates

    geometry = sk.Geometry1D(
        0.6,
        0,
        6372000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.PlaneParallel,
    )

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geos[-1],
            "atmosphere": sk.test_util.scenarios.default_pure_scattering_atmosphere(
                config, geometry, 0.8
            ),
        }
    )

    return scen


def test_wf_extinction():
    """
    Verifies that the extinction derivative is accurate to 5 decimal places
    """
    D_FRACTION = 1e-4
    test_scens = _raw_scenarios()

    for scen in test_scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = engine.calculate_radiance(scen["atmosphere"])

        numeric_wf = np.zeros_like(radiance["wf_extinction"].values)

        for i in range(len(radiance["altitude"])):
            dk = D_FRACTION * scen["atmosphere"].storage.total_extinction[i]

            scen["atmosphere"].storage.total_extinction[i] += dk
            radiance_above = engine.calculate_radiance(scen["atmosphere"])

            scen["atmosphere"].storage.total_extinction[i] -= 2 * dk
            radiance_below = engine.calculate_radiance(scen["atmosphere"])

            scen["atmosphere"].storage.total_extinction[i] += dk

            numeric_wf[:, :, :, i] = (
                radiance_above["radiance"].to_numpy()
                - radiance_below["radiance"].to_numpy()
            ) / (2 * dk)

        radiance["wf_extinction_numeric"] = (
            ["wavelength", "los", "stokes", "altitude"],
            numeric_wf,
        )
        validate_wf(
            radiance["wf_extinction"], radiance["wf_extinction_numeric"], decimal=5
        )


def test_wf_ssa():
    """
    Verifies that the single scatter albedo derivative is accurate to 6 decimal places
    """
    D_FRACTION = 1e-5
    test_scens = _raw_scenarios()

    for scen in test_scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = engine.calculate_radiance(scen["atmosphere"])

        numeric_wf = np.zeros_like(radiance["wf_ssa"].values)

        for i in range(len(radiance["altitude"])):
            d_ssa = D_FRACTION * scen["atmosphere"].storage.ssa[i]

            scen["atmosphere"].storage.ssa[i] += d_ssa
            radiance_above = engine.calculate_radiance(scen["atmosphere"])

            scen["atmosphere"].storage.ssa[i] -= 2 * d_ssa
            radiance_below = engine.calculate_radiance(scen["atmosphere"])

            scen["atmosphere"].storage.ssa[i] += d_ssa

            numeric_wf[:, :, :, i] = (
                radiance_above["radiance"].to_numpy()
                - radiance_below["radiance"].to_numpy()
            ) / (2 * d_ssa)

        radiance["wf_ssa_numeric"] = (
            ["wavelength", "los", "stokes", "altitude"],
            numeric_wf,
        )
        validate_wf(radiance["wf_ssa"], radiance["wf_ssa_numeric"], decimal=6)


def test_wf_legendre():
    """
    Verifies that the legendre coefficient derivatives are accurate to some decimal places
    """
    D_L = 1e-4
    test_scens = _raw_scenarios()

    for scen in test_scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = engine.calculate_radiance(scen["atmosphere"])

        for i in range(scen["atmosphere"].storage.leg_coeff.shape[0]):
            if i == 0:
                continue

            numeric_wf = np.zeros_like(radiance[f"wf_leg_coeff_{i}"].values)

            for j in range(len(radiance["altitude"])):
                scen["atmosphere"].storage.leg_coeff[i, j] += D_L
                radiance_above = engine.calculate_radiance(scen["atmosphere"])

                scen["atmosphere"].storage.leg_coeff[i, j] -= 2 * D_L
                radiance_below = engine.calculate_radiance(scen["atmosphere"])

                scen["atmosphere"].storage.leg_coeff[i, j] += D_L

                numeric_wf[:, :, :, j] = (
                    radiance_above["radiance"].to_numpy()
                    - radiance_below["radiance"].to_numpy()
                ) / (2 * D_L)

            radiance[f"wf_leg_coeff_{i}_numeric"] = (
                ["wavelength", "los", "stokes", "altitude"],
                numeric_wf,
            )
            validate_wf(
                radiance[f"wf_leg_coeff_{i}"],
                radiance[f"wf_leg_coeff_{i}_numeric"],
                decimal=6,
            )



# ruff: noqa: F401

from ._core import (  # noqa: I001
    AtmosphereStokes_1,
    AtmosphereStokes_3,
    AtmosphereStorageStokes_1,
    AtmosphereStorageStokes_3,
    Config,
    EngineStokes_1,
    EngineStokes_3,
    Geodetic,
    Geometry1D,
    GeometryType,
    GroundViewingSolar,
    InterpolationMethod,
    MultipleScatterSource,
    SingleScatterSource,
    OccultationSource,
    OutputIdealStokes_1,
    OutputIdealStokes_3,
    StokesBasis,
    Surface,
    TangentAltitudeSolar,
    ThreadingModel,
    ViewingGeometry,
    ViewingGeometryBase,
)
from ._version import __version__
from .atmosphere import Atmosphere
from .output import Output, OutputIdeal
from .engine import Engine
from .geodetic import WGS84, SphericalGeoid

from . import appconfig, climatology, constituent, mie, optical, test_util, util



import abc

from sasktran2 import Atmosphere


class Constituent(abc.ABC):
    @abc.abstractmethod
    def add_to_atmosphere(self, atmo: Atmosphere):
        pass

    @abc.abstractmethod
    def register_derivative(self, atmo: Atmosphere, name: str):
        pass
